#include <sys/types.h>
#include <sys/time.h>
#include <sys/select.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <lx_string.h>
#include <minimisc.h>
#include <get_opts.h>
#include "defines.h"
#include "sockdata.h"
#include "opt_misc.h"
#include "winch.h"


int maxfd = 0;
fd_set fdset_out,
       fdset_in;

static struct itimerval *
check_timer_next (timeout1, timeout2)
	struct itimerval *timeout1, *timeout2;
{
	struct timeval   *timer1 = &timeout1->it_value,
		         *timer2 = &timeout2->it_value;

	if (!timerisset(timer1) && !timerisset(timer2)) return 0;
	if (!timerisset(timer1)) return timeout2;
	if (!timerisset(timer2)) return timeout1;

	if (timercmp(timer1, timer2, <=)) return timeout1;
	else                              return timeout2;

}


static void
check_socketlink (socklink, incomplete, set)
	struct socketlink *socklink;
	unsigned char *incomplete;
	fd_set *set;
{
	struct socketlink *thislink;

	*incomplete = 0;

	for (thislink = socklink; thislink; thislink = thislink->next) {
		if ((thislink->sd = sock_open (thislink)) == -1)
			*incomplete = 1;
		else {
			if (thislink->sd > maxfd) maxfd = thislink->sd;
			FD_SET(thislink->sd, set);
		}

		sock_alloc_buf(thislink);
	}

}

static void
alrm_handler (int i)
{
	static struct itimerval ztimeout =
		{ {0,0}, {0,0} };

	dprintf("alarm\n");
	setitimer (ITIMER_REAL, &ztimeout, 0);
	sig_default_alrm();
}


#define  SECS(mil)  ((mil) / 1000)
#define USECS(mil) (((mil) % 1000) * 1000)

void
select_loop (void)
{
	unsigned char incomplete_out = 1,
	              incomplete_in  = 1,
		      pending[2];

	int rv;

	fd_set *fdsp_out, *fdsp_in,
	        fdset_o2,  fdset_i2;

	struct itimerval otimeout, itimeout,
	                *timer_next = (void *)-1; /* nonzero */


	/* output file timeout */
	otimeout.it_interval.tv_sec  = SECS(OPT_RETRY_OTIMEOUT);
	otimeout.it_interval.tv_usec = USECS(OPT_RETRY_OTIMEOUT);
	otimeout.it_value.tv_sec     = otimeout.it_interval.tv_sec;
	otimeout.it_value.tv_usec    = otimeout.it_interval.tv_usec;

	/* input file timeout */
	itimeout.it_interval.tv_sec  = SECS(OPT_RETRY_ITIMEOUT);
	itimeout.it_interval.tv_usec = USECS(OPT_RETRY_ITIMEOUT);
	itimeout.it_value.tv_sec     = itimeout.it_interval.tv_sec;
	itimeout.it_value.tv_usec    = itimeout.it_interval.tv_usec;


	FD_ZERO(&fdset_out);
	FD_ZERO(&fdset_in);

	pending[SW_OUTPUT] = 0;
	pending[SW_INPUT]  = 1;

	dprintf ("Entering select() loop.\n");
	while (1) {
		if (incomplete_out) {
			check_socketlink(sl_out, &incomplete_out, &fdset_out);
			if (!incomplete_out) timerclear(&otimeout.it_value);
		}

		if (incomplete_in) {
			check_socketlink(sl_in,  &incomplete_in, &fdset_in);
			if (!incomplete_in) timerclear(&itimeout.it_value);
		}

		if (timer_next)
			timer_next = check_timer_next(&itimeout, &otimeout);

		if (timer_next) {
			dprintf ("setting new timer.\n");
			sig_catch_alrm(alrm_handler);
			setitimer (ITIMER_REAL, timer_next, 0);
		}

		/* don't read any data unless the outbound socket is open */
		if (incomplete_out) {
			dprintf ("sleeping: awaiting alarm...\n");
			select (0, 0, 0, 0, 0);
#if 0
			timeradd(&timer_next->it_value,
				 &timer_next->it_interval,
				 &timer_next->it_value);
#endif
			continue;
		}

		/* linux may modify descriptor sets */
		memcpy (&fdset_o2, &fdset_out, sizeof(fd_set));
		memcpy (&fdset_i2, &fdset_in,  sizeof(fd_set));

		if (pending[SW_OUTPUT]) dprintf ("pending output\n");
		if (pending[SW_INPUT])  dprintf ("pending input\n");

		if (pending[SW_OUTPUT]) fdsp_out = &fdset_o2;
		else                    fdsp_out = 0;

		if (pending[SW_INPUT])  fdsp_in  = &fdset_i2;
		else                    fdsp_in  = 0;

		dprintf ("waiting for descriptors (max %d).\n", maxfd+1);
		rv = select (maxfd+1, fdsp_in, fdsp_out, 0, 0);

		if (rv == -1) {
			if (errno == EINTR) {
				dprintf ("alarm caught.\n");
				/* XXX: anything else? */
				continue;
			}
			die_error1("select");
		}

		dprintf ("got descriptors.\n");
		xfer_data (rv, fdsp_out, sl_out,
			       fdsp_in,  sl_in, pending);


	}
}
