#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <lx_string.h>
#include <get_opts.h>
#include "defines.h"
#include "sockdata.h"
#include "opt_misc.h"
#include "winch.h"


/***************************************************************************
  sock_act (lx_s *name, int type, int rm, int (*func)())

  Generic function for bind() and connect() calls.  Created because of the
  similarities in the handling of these two functions.  If (rm), then 'name'
  is removed before the action takes place.

  XXX: Currently, there is no provision for connecting to a socket of the
  wrong suggested type.  To correct: first, discover what error will be
  returned (the manual page is not helpful here); verify that this error is
  uniform across platforms.  Then, account for it by trying the other type.

  IMPORTANT: 'name' must be null-terminated beforehand.

  returns the new socket descriptor, or -1 on error.

 ***************************************************************************/

#define V_CLOSE(fd) do { \
			int tmperr = errno; \
			while (close(sd))  \
				if (errno != EINTR) break; \
			errno = tmperr; \
		} while (0)

static int
sock_act (name, type, rm, func)
	lx_s *name;
	int type;
	int rm;
	int (*func)(int, struct sockaddr *, socklen_t);
{
	int sd, size;
	struct sockaddr_un addr;

	sd = socket (AF_LOCAL, type, 0);
	if (sd == -1) return -1;

	memset (&addr, 0, sizeof(addr));
	addr.sun_family = AF_LOCAL;

	/* name->len contains terminating 0 */
	size = sizeof(addr.sun_path);

	if (name->len > size) {
		errno = ENAMETOOLONG;
		return -1;
	}

	memcpy (addr.sun_path, name->s, name->len);

	if (rm) (void) unlink (name->s);

	if (func (sd, (struct sockaddr *)&addr,
			sizeof(addr) - (size - name->len))) {
		V_CLOSE(sd);
		return -1;
	}

	if (set_socket_options (sd)) {
		V_CLOSE(sd);
		return -1;
	}

	dprintf ("%s: succeeded (descriptor %d).\n", name->s, sd);
	return sd;

}

/***************************************************************************
  sock_bind (lx_s *name, int type)

  Removes 'name', then binds it to a socket of type 'type'.
  If the socket is of type SOCK_STREAM, calls listen(2) as well.

  IMPORTANT: 'name' must be null-terminated beforehand.

  returns the new socket descriptor, or -1 on error.

 ***************************************************************************/

int
sock_bind (name, type)
	lx_s *name;
	int type;
{
	int sd;

	dprintf ("attempting bind of %s\n", name->s);

	sd = sock_act (name, type, 1, bind);
	if (sd == -1) return -1;

	if (type == SOCK_STREAM) {
		dprintf ("STREAM: calling listen on descriptor %d\n", sd);
		if (listen (sd, 4)) {
			V_CLOSE(sd);
			return -1;
		}
	}

	return sd;
}

/***************************************************************************
  sock_connect (lx_s *name, int type)

  connects to 'name', first trying socket of type 'type'.

  IMPORTANT: 'name' must be null-terminated beforehand.

  returns the connected socket descriptor, or -1 on error.

 ***************************************************************************/

int
sock_connect (name, type)
	lx_s *name;
	int type;
{
	dprintf ("attempting connect to %s\n", name->s);
	return (sock_act (name, type, 0, connect));
}

/***************************************************************************
  sock_open (struct socketlink *sl)

  Attempts to open the socket to be associated with sl->name.  If the BIND
  option has been given, it will attempt to remove and recreate the socket.
  Otherwise, it will try a simple connect(), taking a hint from the relevant
  socket type.

  returns the newly created socket, or -1 on error.

 ***************************************************************************/
int
sock_open (sl)
	struct socketlink *sl;
{
	lx_check0(&sl->name);

	if (sl->method == SW_OUTPUT) {
		sl->type = socktype_out;
	       	if (OPT_BIND_OUTSOCK) {
			return (sock_bind(&sl->name, sl->type));
		}
	}

	if (sl->method == SW_INPUT) {
		sl->type = socktype_in;
	       	if (OPT_BIND_SOCKS) {
			return (sock_bind(&sl->name, sl->type));
		}
	}

	return (sock_connect(&sl->name, sl->type));
}
