#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <string.h>
#include <lx_string.h>
#include <get_opts.h>
#include "defines.h"
#include "sockdata.h"
#include "winch.h"


/***************************************************************************
  write_data (int *nready, fd_set *set_out, struct socketlink *outsocks,
                                            unsigned char *pending);

  Writes output buffer(s) to the respective socket(s).
  Sets pending, depending on whether anything is left in the output buffer.
  Decrements *nready as necessary.

  returns nothing.

 ***************************************************************************/

static inline void
write_data (nready, set_out, outsocks, pending)
	int *nready;
	fd_set *set_out;
	struct socketlink *outsocks;
	unsigned char *pending;
{
	int rv;

	dprintf ("   write\n");
	for (; outsocks && *nready; outsocks = outsocks->next) {

		dprintf ("checking descriptor %d\n", outsocks->sd);
		if (outsocks->sd == -1)
			continue;

		if (!FD_ISSET(outsocks->sd, set_out))
			continue;

		dprintf ("      set: %s\n", outsocks->name.s);
		(*nready)--;

		if (!outsocks->buf->len)
			continue;

		rv = sendto(outsocks->sd,
			    outsocks->buf->buf,
			    outsocks->buf->len, 0, 0, 0);

		if (rv == -1)
			die_error1("sendto");

		if (outsocks->buf->len -= rv) {
			memmove (outsocks->buf->buf,
				 outsocks->buf->buf + rv,
				 outsocks->buf->len);
			pending[SW_OUTPUT] = 1;
		} else  pending[SW_OUTPUT] = 0;

		dprintf ("         %d bytes of data sent.\n", rv);
		dprintf ("         %d bytes of data waiting.\n",
				outsocks->buf->len);

		if (outsocks->buf->len < outsocks->buf->max)
			pending[SW_INPUT] = 1;
		else    pending[SW_INPUT] = 0;
	}
}


/***************************************************************************
  read_data (int *nready, fd_set *set_in, struct socketlink *insocks,
                                          unsigned char *pending);

  Reads data from sockets; appends to output buffer(s).
  Sets pend_in if there is still space available in the output buffer.
  Decrements *nready as necessary.

  returns nothing.

 ***************************************************************************/

extern fd_set fdset_in;
extern struct socketlink *sl_in;

static inline void
read_data (nready, set_in, outsocks, insocks, lastsock, pending)
	int *nready;
	fd_set *set_in;
	struct socketlink *outsocks, *insocks, *lastsock;
	unsigned char *pending;
{
	struct socketlink *thissock = insocks,
	                  *prevsock = 0;
	int rv;

	dprintf ("   read\n");
	for ( ; thissock && *nready; ) {

		dprintf ("checking descriptor %d\n", thissock->sd);
		if (thissock->sd == -1)
			goto CONTINUE;

		if (!FD_ISSET(thissock->sd, set_in))
			goto CONTINUE;

		(*nready)--;

		if (thissock->type == SOCK_STREAM) {
			socket_accept(&fdset_in, thissock);
			goto CONTINUE;
		}


		dprintf ("      set: %s\n", thissock->name.s);

		if (outsocks->buf->max == outsocks->buf->len)
			return;

		rv = recvfrom(thissock->sd,
			      outsocks->buf->buf + outsocks->buf->len,
			      outsocks->buf->max - outsocks->buf->len,
			      0, 0, 0);

		if (rv == -1)
			die_error1("recvfrom");

		if (!rv && (thissock->type == SOCK_ESTABLISHED)) {
			struct socketlink **sockp;
			if (!prevsock) sockp = &sl_in;
			else           sockp = &(prevsock->next);

			thissock = thissock->next;
			socket_established_close(&fdset_in, sockp);
			/* prevsock stays the same */
			continue;
		}

		if (outsocks->buf->len += rv) pending[SW_OUTPUT] = 1;
		else                          pending[SW_OUTPUT] = 0;

		dprintf ("         %d bytes of data read.\n", rv);
		dprintf ("         %d bytes of data available.\n",
				outsocks->buf->max - outsocks->buf->len);

		if (outsocks->buf->len < outsocks->buf->max)
			pending[SW_INPUT] = 1;
		else {
			pending[SW_INPUT] = 0;
			lastsock = thissock;
			return;
		}

CONTINUE:
		prevsock = thissock;
		thissock = thissock->next;
	}

	lastsock = 0;
}

void
xfer_data (nready, set_out, outsocks, set_in, insocks, pending)
	int nready;
	unsigned char *pending;
	fd_set *set_out,
	       *set_in;
	struct socketlink *outsocks,
	                  *insocks;
{
	static struct socketlink *lastsock = 0;

	dprintf ("transferring data:\n");
	if (!nready) return;
	if (pending[SW_OUTPUT])
		write_data (&nready, set_out, outsocks, pending);

	if (!nready) return;
	if (pending[SW_INPUT]) {
		if (!lastsock) lastsock = insocks;
		read_data  (&nready, set_in, outsocks, insocks,
				             lastsock, pending);
	}
}
